.opview {
	display: none;
}
.opview.act {
	display: block;
}
#ops a {
	color: #fc5;
	font-size: 1.5em;
	padding: 0 .3em;
	margin: 0;
	outline: none;
}
#ops a.act {
	text-decoration: underline;
}
/*
#ops a+a:after,
#ops a:first-child:after {
	content: 'x';
	color: #282828;
	text-shadow: 0 0 .08em #01a7e1;
	margin-left: .3em;
	position: relative;
}
#ops a+a:before {
	content: 'x';
	color: #282828;
	text-shadow: 0 0 .08em #ff3f1a;
	margin-right: .3em;
	margin-left: -.3em;
}
#ops a:last-child:after {
	content: '';
}
#ops a.act:before,
#ops a.act:after {
	text-decoration: none !important;
}
*/
#ops i {
	font-size: 1.5em;
}
#ops i:before {
	content: 'x';
	color: #282828;
	text-shadow: 0 0 .08em #01a7e1;
	position: relative;
}
#ops i:after {
	content: 'x';
	color: #282828;
	text-shadow: 0 0 .08em #ff3f1a;
	margin-left: -.35em;
	font-size: 1.05em;
}
#ops,
.opbox {
	border: 1px solid #3a3a3a;
	box-shadow: 0 0 1em #222 inset;
}
#ops {
	display: none;
	background: #333;
	margin: 1.7em 1.5em 0 1.5em;
	padding: .3em .6em;
	border-radius: .3em;
	border-width: .15em 0;
}
.opbox {
	background: #2d2d2d;
	margin: 1.5em 0 0 0;
	padding: .5em;
	border-radius: 0 1em 1em 0;
	border-width: .15em .3em .3em 0;
	max-width: 40em;
}
.opbox input {
	margin: .5em;
}
.opbox input[type=text] {
	color: #fff;
	background: #383838;
	border: none;
	box-shadow: 0 0 .3em #222;
	border-bottom: 1px solid #fc5;
	border-radius: .2em;
	padding: .2em .3em;
}
#op_up2k {
	padding: 0 1em 1em 1em;
}
#u2form {
	position: absolute;
	top: 0;
	left: 0;
}
#u2form input {
	background: #444;
	border: 0px solid #444;
	outline: none;
}
#u2err.err {
	color: #f87;
	padding: .5em;
}
#u2form {
	width: 2px;
	height: 2px;
	overflow: hidden;
}
#u2btn {
	color: #eee;
	background: #555;
	background: -moz-linear-gradient(top, #367 0%, #489 50%, #38788a 51%, #367 100%);
	background: -webkit-linear-gradient(top, #367 0%, #489 50%, #38788a 51%, #367 100%);
	background: linear-gradient(to bottom, #367 0%, #489 50%, #38788a 51%, #367 100%);
	filter: progid:DXImageTransform.Microsoft.gradient(startColorstr='#489', endColorstr='#38788a', GradientType=0);
	text-decoration: none;
	line-height: 1.5em;
	border: 1px solid #222;
	border-radius: .4em;
	text-align: center;
	font-size: 2em;
	margin: 1em auto;
	padding: 1em 0;
	width: 12em;
	cursor: pointer;
	box-shadow: .4em .4em 0 #111;
}
#u2notbtn {
	display: none;
	text-align: center;
	background: #333;
	padding-top: 1em;
}
#u2notbtn * {
	line-height: 1.3em;
}
#u2tab {
	margin: 3em auto;
	width: calc(100% - 2em);
	max-width: 100em;
}
#u2tab td {
	border: 1px solid #ccc;
	border-width: 0 0px 1px 0;
	padding: .1em .3em;
}
#u2tab td:nth-child(2) {
	width: 5em;
}
#u2tab td:nth-child(3) {
	width: 40%;
}
#u2tab tr+tr:hover td {
	background: #222;
}
#u2conf {
	margin: 1em auto;
	width: 26em;
}
#u2conf * {
	text-align: center;
	line-height: 1em;
	margin: 0;
	padding: 0;
	border: none;
	outline: none;
}
#u2conf .txtbox {
	width: 4em;
	color: #fff;
	background: #444;
	border: 1px solid #777;
	font-size: 1.2em;
	padding: .15em 0;
}
#u2conf a {
	color: #fff;
	background: #c38;
	text-decoration: none;
	border-radius: .1em;
	font-size: 1.5em;
	padding: .1em 0;
	margin: 0 -.25em;
	width: 1.5em;
	height: 1em;
	display: inline-block;
	position: relative;
	line-height: 1em;
	bottom: -.08em;
}
#u2conf input+a {
	background: #d80;
}
#u2foot {
	color: #fff;
	font-style: italic;
}
.prog {
	font-family: monospace;
}
.prog>div {
	display: inline-block;
	position: relative;
	overflow: hidden;
	margin: 0;
	padding: 0;
	height: 1.1em;
	margin-bottom: -.15em;
	box-shadow: -1px -1px 0 inset rgba(255,255,255,0.1);
}
.prog>div>div {
	width: 0%;
	position: absolute;
	left: 0;
	top: 0;
	bottom: 0;
	background: #0a0;
}
